class NotifyType:
    INFO: NotifyType
    SUCCESS: NotifyType
    WARNING: NotifyType
    FAILURE: NotifyType

class NotifyFormat:
    TEXT: NotifyFormat
    HTML: NotifyFormat
    MARKDOWN: NotifyFormat

class ContentLocation:
    LOCAL: ContentLocation
    HOSTED: ContentLocation
    INACCESSIBLE: ContentLocation